/*********************************************************************

          ps.c: PolhemusSensor device driver

                                        Kenji with Tail

*********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termio.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stropts.h>
#include <unistd.h>

#include "ps.h"


#define _DEBUG_ 1


/*********          Serial Port comunication driver          ********/
/* Public constants for IRIX... */
#define OFLAGS	(OPOST | OLCUC | ONLCR | OCRNL | ONOCR  \
			| ONLRET | OFILL | OFDEL)
#define LFLAGS	(ICANON | ISIG | XCASE | ECHO | ECHOE | ECHOK  \
			| ECHONL | NOFLSH)
#define IFLAGS	(IGNBRK | BRKINT | IGNPAR | PARMRK | INPCK | INLCR  \
			| ICRNL | IGNCR | IUCLC | ISTRIP | IXON  \
			| IXOFF)

#define MIN_TIMEOUT 1
#define TIM_TIMEOUT 0


#define CTL_COM(n) ((n) - 0x40)
#define CR_COM 13		/* դä: 礦Ȥ */
#define LF_COM 10		/* 礦: ĤΤ礦 */
#define NUL_COM 0		/* NULL ⤸ */


#define NUM_COM	(4+1)	/* port is 4, 0-4 */

static	int		port_ref = -1;
static	struct	termio	old_setup;


/* OpenSP_com() opens the given serial port with specified baud rate
 *    Always uses 8 data bits, 1 stop bit, no parity.
 *    Assumes port number is valid and baud rate has been fixed up.
 *    Returns False (zero) if can't open the port.
 */
static int OpenSP_com(char *port_name, long int baud)
{
	int		result = 0;	/* Assume the port will not be accessible */
	tcflag_t	baud_code;
	struct	termio	new_setup;

	if (port_ref != -1){
		fprintf(stderr, "serial port has been opened");
		return result;
	}

	port_ref = open(port_name, O_RDWR);

	/* If port and its parameters are accessible, reconfigure for Probe */
	if ((port_ref >= 0)
	    && (ioctl(port_ref, TCGETA, &old_setup) >= 0)){

		/* Save original set-up */
		new_setup = old_setup;

		/* Turn off all post-processing and char translation */
		new_setup.c_oflag &= ~OFLAGS;
		new_setup.c_lflag &= ~LFLAGS;
		new_setup.c_iflag &= ~IFLAGS;

		/* Set for no waiting during char read */
		new_setup.c_cc[VMIN] = MIN_TIMEOUT;
		new_setup.c_cc[VTIME] = TIM_TIMEOUT;

		/* Find appropriate baud rate flags */
		switch(baud){
		    case 0L:
			baud_code = B0;		break; /* down ER signal, close */
		    case 50L:
			baud_code = B50;	break;
		    case 75L:
			baud_code = B75;	break;
		    case 110L:
			baud_code = B110;	break;
		    case 134L:
			baud_code = B134;	break;
		    case 150L:
			baud_code = B150;	break;
		    case 200L:
			baud_code = B200;	break;
		    case 300L:
			baud_code = B300;	break;
		    case 600L:
			baud_code = B600;	break;
		    case 1200L:
			baud_code = B1200;	break;
		    case 1800L:
			baud_code = B1800;	break;
		    case 2400L:
			baud_code = B2400;	break;
		    case 4800L:
			baud_code = B4800;	break;
		    case 9600L:
			baud_code = B9600;	break;
		    case 19200L:
			baud_code = B19200;	break;
		    case 38400L:
			baud_code = B38400;	break;
		    case 'A':
			baud_code = EXTA;	break;
		    case 'B':
			baud_code = EXTB;	break;
		    default:
			fprintf(stderr, "OpenSP_com: bad baud rate: %ld\n", baud);
		}

		/* Set for baud rate, 8 data bits, 1 stop bit, no parity */
		new_setup.c_cflag = baud_code | CS8 | CREAD;

		/* Set these new parameters.  Record success if it works. */
		if (ioctl(port_ref, TCSETA, &new_setup) >= 0){
			result = 1;
		}
	}

	return result;

}


/********************************************************************/


/* station#1-4 active flag */
static	int	stn_flg[5] = {OFF_PS, OFF_PS, OFF_PS, OFF_PS, OFF_PS};

/* dummy for PsCgShm */
char 	 	_ps_server4shm_[256];

/* SetSaveDataSpl() */
int SetSaveDataSpl(void);

/* ChangeUnit_ps() changes unit inchi/meter
 *    Returns False (zero) if buffer is full
 *    Returns True (non-zero) if successful
 *    give unit:  inchi: INCHI_PS/ meter: METER_PS.
 */
#define INCHI_PS 0
#define METER_PS 1
int ChangeUnit_ps(int unit);

/* ChangeFormat_ps() changes format ascii/binary
 *    Returns False (zero) if buffer is full
 *    Returns True (non-zero) if successful
 *    give form:  ascii: ASCII_PS/ binary: BINARY_PS.
 */
#define ASCII_PS 0
#define BINARY_PS 1
int ChangeFormat_ps(int form);

/* ChangeCompens_ps() changes metal compensation enable/disable
 *    Returns False (zero) if buffer is full
 *    Returns True (non-zero) if successful
 *    give fixed:  enable: ENABLE_PS/ disable: DISABLE_PS.
 */
#define DISABLE_PS 0
#define ENABLE_PS 1
int ChangeCompens_ps(int fixed);

/* StylusButton_ps() sets button of stylus of station 1 for MouseMode
 *    Returns False (zero) if buffer is full
 *    Returns True (non-zero) if successful
 */
int StylusButton_ps(void);

/* PosRequestA_ps() requests ascii position and orient
 *    Returns 1 if station#1 is stylus and button is ON
 *    Returns 0 other else
 *    give p*:    position/orient structur.
 */
int PosRequestA_ps(PosOri_ps p1, PosOri_ps p2, PosOri_ps p3, PosOri_ps p4);


int Init_ps(char *port, long int baud)
{
	int result;

	result = OpenSP_com(port, baud);

	if(result == 1){
		ioctl(port_ref, TCFLSH, 2); /* 2 means flush both buffers */

		Reset_ps();
		StylusButton_ps();
		ChangeFormat_ps(BINARY_PS);
		ChangeUnit_ps(METER_PS);
		ChangeCompens_ps(ENABLE_PS); /* comp. data must be in EPROM? */
		SetStation_ps(1, ON_PS);
		SetStation_ps(2, OFF_PS);
		SetStation_ps(3, OFF_PS);
		SetStation_ps(4, OFF_PS);
	}

	return result;
}


int Close_ps(void)
{
	if(port_ref != -1){
		ioctl(port_ref, TCSETA, &old_setup);
		close(port_ref);
		port_ref = -1;
		return 1;
	} else {
		fprintf(stderr, "ps has not been opened\n");
		return 0;
	}
}


int Reset_ps(void)
{
	char c = CTL_COM('Y');
	int result, i;

	result = (write(port_ref, &c, 1) == 1);
	fprintf(stderr, "Reset_ps: now resetting ps, please wait 15 sec");
	/* sleep(15); */
	for(i = 0; i < 15; i ++){
		sleep(1);
		fprintf(stderr, ".");
	}
	fprintf(stderr, "done.\n");

	return result;
}


int ChangeUnit_ps(int unit)
{
	switch(unit){
	    case INCHI_PS:
		return (write(port_ref, "U", 1) == 1);
	    case METER_PS:
		return (write(port_ref, "u", 1) == 1);
	    default:
		fprintf(stderr, "ChangeUnit_ps: please give INCHI_PS/METER_PS.\n");
		return -1;
	}
}


int ChangeFormat_ps(int form)
{
	switch(form){
	    case ASCII_PS:
		return (write(port_ref, "F", 1) == 1);
	    case BINARY_PS:
		return (write(port_ref, "f", 1) == 1);
	    default:
		fprintf(stderr, "ChangeFormat_ps: please give ASCII_PS/BINARY_PS.\n");
		return -1;
	}
}


int ChangeCompens_ps(int fixed)
{
	switch(fixed){
	    case ENABLE_PS:
		return (write(port_ref, "D", 1) == 1);
	    case DISABLE_PS:
		return (write(port_ref, "d", 1) == 1);
	    default:
		fprintf(stderr, "ChangeCompens_ps: please give ENABLE_PS/DISABLE_PS.\n");
		return -1;
	}
}


int StylusButton_ps(void)
{
	char str[] = "e1,0\r";

	return (write(port_ref, str, 5) == 5);
}


int SetStation_ps(int stn, int act)
{
	int result;
	char actstr[] = "l0,0\r";
	char frmstr[] = "O0,2,4\r";
	char stystr[] = "O1,2,4,16\r";

	actstr[1] += stn;
	actstr[3] += act;
	stn_flg[stn] = act;

	result = (write(port_ref, actstr, 5) == 5); /* set activation */

	if(act == OFF_PS) return result;

	if(stn == 1){
		return (write(port_ref, stystr, 10) == 10); /* set stylus data */
	} else {
		frmstr[1] += stn;
		return (write(port_ref, frmstr, 7) == 7); /* set normal data */
	}
}



/* GetSty_ps() gets position, orient and stylus button
 *    Returns stylus button is ON/OFF
 *    give p:     position/orient structur.
 */
static int GetSty_ps(PosOri_ps p)
{
	int i, j;
	char stn, err, c, *f;

	f = (char *) p;

	read(port_ref, &c, 1);
	read(port_ref, &stn, 1); /* station number */
	read(port_ref, &err, 1); /* system error code */
	stn = stn - '0';
#if _DEBUG_
	if(c != '0' || stn != 1){
		fprintf(stderr, "GetSty_ps-1: %d.\n", c);
		fprintf(stderr, "GetSty_ps-ST: %d.\n", stn);
		fprintf(stderr, "GetSty_ps-EC: %c.\n", err);
	}
#endif

	for(i = 0; i < 6; i++){
		for(j = 0; j < 4; j++){
			read(port_ref, f + (4 * i  + 3 - j), 1);
		}
	}

	read(port_ref, &c, 1);
#if _DEBUG_
	if(c != ' '){
		fprintf(stderr, "GetSty_ps-2: %d.\n", c);
	}
#endif
	read(port_ref, &c, 1);

	return (int) (c - '0');
}


/* GetPos_ps() gets position and orient
 *    Returns station number
 *    give p:     position/orient structur.
 */
static int GetPos_ps(int station, PosOri_ps p)
{
	int i, j;
	char stn, err, c, *f;

	f = (char *) p;

	read(port_ref, &c, 1);
	read(port_ref, &stn, 1); /* station number */
	read(port_ref, &err, 1); /* system error code */
	stn = stn - '0';
#if _DEBUG_
	if(c != '0' || stn != station){
		fprintf(stderr, "GetPos_ps(%d)-1: %d.\n", station, c);
		fprintf(stderr, "GetPos_ps(%d)-ST: %d.\n", station, stn);
		fprintf(stderr, "GetPos_ps(%d)-EC: %c.\n", station, err);
	}
#endif

	for(i = 0; i < 6; i++){
		for(j = 0; j < 4; j++){
			read(port_ref, f + (4 * i  + 3 - j), 1);
		}
	}

	return stn;
}


/* GetPosA_ps() gets position and orient
 *    Returns station number
 *    Returns 0 if station is 1 and stylus button is ON
 *    give p:     position/orient structur.
 */
static int GetPosA_ps(int station, PosOri_ps p)
{
	int i, j;
	char stn, err, c, a[8];

	read(port_ref, &c, 1);
	read(port_ref, &stn, 1);
	read(port_ref, &err, 1); /* system error code */
	stn = stn - '0';
#if _DEBUG_
	if(c != '0' || stn != station){
		fprintf(stderr, "GetPosA_ps(%d)-1: %d.\n", station, c);
		fprintf(stderr, "GetPosA_ps(%d)-ST: %d.\n", station, stn);
		fprintf(stderr, "GetPosA_ps(%d)-EC: %c.\n", station, err);
	}
#endif

	for(i = 0; i < 6; i++){
		for(j = 0; j < 7; j++){
			read(port_ref, a + j, 1);
		}
		a[j] = '\0';
		p[i] = (float) atof(a);
	}

	if(stn != 1) return stn;

	read(port_ref, &c, 1);
#if _DEBUG_
	if(c != ' '){
		fprintf(stderr, "GetPosA_ps(%d)-2: %d.\n", station, c);
	}
#endif
	read(port_ref, &c, 1);

	return (int) (c - '0');
}


int PosRequest_ps(PosOri_ps p1, PosOri_ps p2, PosOri_ps p3, PosOri_ps p4)
{
	int btn = 0;

	write(port_ref, "P", 1);

	if(stn_flg[1] == ON_PS){
		btn = GetSty_ps(p1);
	}
	if(stn_flg[2] == ON_PS){
		GetPos_ps(2, p2);
	}
	if(stn_flg[3] == ON_PS){
		GetPos_ps(3, p3);
	}
	if(stn_flg[4] == ON_PS){
		GetPos_ps(4, p4);
	}

	ioctl(port_ref, TCFLSH, 2); /* 2 means flush both buffers */

	return btn;
}


int PosRequestA_ps(PosOri_ps p1, PosOri_ps p2, PosOri_ps p3, PosOri_ps p4)
{
	int btn = 0;

	write(port_ref, "P", 1);

	if(stn_flg[1] == ON_PS){
		btn = GetPosA_ps(1, p1);
	}
	if(stn_flg[2] == ON_PS){
		GetPosA_ps(2, p2);
	}
	if(stn_flg[3] == ON_PS){
		GetPosA_ps(3, p3);
	}
	if(stn_flg[4] == ON_PS){
		GetPosA_ps(4, p4);
	}

	ioctl(port_ref, TCFLSH, 2); /* 2 means flush both buffers */

	return btn;
}


/* T_T T_T T_T T_T T_T T_T T_T T_T T_T T_T */

/* TxMountFrame_ps() */
int TxMountFrame_ps(int stn, float a, float e, float r);

/* AttitudeFilter_ps() */
int AttitudeFilter_ps(float f, float flow, float fhigh, float factor);

/* PositionFilter_ps() */
int PositionFilter_ps(float f, float flow, float fhigh, float factor);

/* SetSynchMode_ps() */
int SetSynchMode_ps(int mode);

/* SeveOpCfg_ps() */
int SeveOpCfg_ps(void);

/* ResetSystemDef_ps() */
int ResetSystemDef_ps(void);


int TxMountFrame_ps(int stn, float a, float e, float r)
{
	int w;
	char str[] = "r0,+000.000,+000.000,+000.000\r\0";

	sprintf(str, "r%d,%-7.3f,%-7.3f,%-7.3f\r\0", stn, a, e, r);

	w = (int) strlen(str);
	return(write(port_ref, str, w) == w);
}


int AttitudeFilter_ps(float f, float flow, float fhigh, float factor)
{
	int w;
	char str[] = "v0, +0.000, +0.000, +0.000, +0.000\r\0";

	sprintf(str, "v%-5.3f,%-5.3f,%-5.3f,%-5.3f\r\0", f, flow, fhigh, factor);

	w = (int) strlen(str);
	return(write(port_ref, str, w) == w);
}


int PositionFilter_ps(float f, float flow, float fhigh, float factor)
{
	int w;
	char str[] = "x0, +0.000, +0.000, +0.000, +0.000\r\0";

	sprintf(str, "x%-5.3f,%-5.3f,%-5.3f,%-5.3f\r\0", f, flow, fhigh, factor);

	w = (int) strlen(str);
	return(write(port_ref, str, w) == w);
}


int SetSynchMode_ps(int mode)
{
	char str[] = "y0\r";

	str[1] += mode;
	return(write(port_ref, str, 3) == 3);
}


int SeveOpCfg_ps(void)
{
	char c = CTL_COM('K');
	int result, i;

	result = (write(port_ref, &c, 1) == 1);
	fprintf(stderr, "SeveOpCfg_ps: now saving data of ps, please wait 5 sec");
	/* sleep(5); */
	for(i = 0; i < 5; i ++){
		sleep(1);
		fprintf(stderr, ".");
	}
	fprintf(stderr, "done.\n");

	return result;
}


int ResetSystemDef_ps(void)
{
	return(write(port_ref, "W", 1) == 1);
}


int SetSaveDataSpl(void)
{
	ResetSystemDef_ps();

	AttitudeFilter_ps(.2, .2, .8, .8);
	PositionFilter_ps(.2, .2, .8, .8);
/*	SetSynchMode_ps(2);*/

	SeveOpCfg_ps();
	return Reset_ps();
}


