/*********************************************************************

          cgs.c: CyberGlove device driver server

                                        Kenji with Tail

*********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <termio.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stropts.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "cg.h"


#define _DEBUG_ 1


/*********          Serial Port comunication driver          ********/
/* Public constants for IRIX... */
#define OFLAGS	(OPOST | OLCUC | ONLCR | OCRNL | ONOCR  \
			| ONLRET | OFILL | OFDEL)
#define LFLAGS	(ICANON | ISIG | XCASE | ECHO | ECHOE | ECHOK  \
			| ECHONL | NOFLSH)
#define IFLAGS	(IGNBRK | BRKINT | IGNPAR | PARMRK | INPCK | INLCR  \
			| ICRNL | IGNCR | IUCLC | ISTRIP | IXON  \
			| IXOFF)

#define MIN_TIMEOUT 1
#define TIM_TIMEOUT 0


#define CTL_COM(n) ((n) - 0x40)
#define CR_COM 13		/* դä: 礦Ȥ */
#define LF_COM 10		/* 礦: ĤΤ礦 */
#define NUL_COM 0		/* NULL ⤸ */


#define NUM_COM	(4+1)	/* port is 4, 0-4 */

static	int		port_ref[NUM_COM] = {-1, -1, -1, -1, -1};
static	struct	termio	old_setup[NUM_COM];


/* OpenSP_com() opens the given serial port with specified baud rate
 *    Always uses 8 data bits, 1 stop bit, no parity.
 *    Assumes port number is valid and baud rate has been fixed up.
 *    Returns False (zero) if can't open the port.
 */
static int OpenSP_com(int port_id, char *port_name, long int baud)
{
	int		result = 0;	/* Assume the port will not be accessible */
	tcflag_t	baud_code;
	struct	termio	new_setup;

	if (port_ref[port_id] != -1){
		fprintf(stderr, "serial port(%d) has been opened", port_id);
		return result;
	}

	port_ref[port_id] = open(port_name, O_RDWR);

	/* If port and its parameters are accessible, reconfigure for Probe */
	if ((port_ref[port_id] >= 0)
	    && (ioctl(port_ref[port_id], TCGETA, &old_setup[port_id]) >= 0)){

		/* Save original set-up */
		new_setup = old_setup[port_id];

		/* Turn off all post-processing and char translation */
		new_setup.c_oflag &= ~OFLAGS;
		new_setup.c_lflag &= ~LFLAGS;
		new_setup.c_iflag &= ~IFLAGS;

		/* Set for no waiting during char read */
		new_setup.c_cc[VMIN] = MIN_TIMEOUT;
		new_setup.c_cc[VTIME] = TIM_TIMEOUT;

		/* Find appropriate baud rate flags */
		switch(baud){
		    case 0L:
			baud_code = B0;		break; /* down ER signal, close */
		    case 50L:
			baud_code = B50;	break;
		    case 75L:
			baud_code = B75;	break;
		    case 110L:
			baud_code = B110;	break;
		    case 134L:
			baud_code = B134;	break;
		    case 150L:
			baud_code = B150;	break;
		    case 200L:
			baud_code = B200;	break;
		    case 300L:
			baud_code = B300;	break;
		    case 600L:
			baud_code = B600;	break;
		    case 1200L:
			baud_code = B1200;	break;
		    case 1800L:
			baud_code = B1800;	break;
		    case 2400L:
			baud_code = B2400;	break;
		    case 4800L:
			baud_code = B4800;	break;
		    case 9600L:
			baud_code = B9600;	break;
		    case 19200L:
			baud_code = B19200;	break;
		    case 38400L:
			baud_code = B38400;	break;
		    case 'A':
			baud_code = EXTA;	break;
		    case 'B':
			baud_code = EXTB;	break;
		    default:
			fprintf(stderr, "OpenSP_com: bad baud rate: %ld\n", baud);
		}

		/* Set for baud rate, 8 data bits, 1 stop bit, no parity */
		new_setup.c_cflag = baud_code | CS8 | CREAD;

		/* Set these new parameters.  Record success if it works. */
		if (ioctl(port_ref[port_id], TCSETA, &new_setup) >= 0){
			result = 1;
		}
	}

	return result;

}


/********************************************************************/


/* cg#1-4 get-light flag */
static	int	lit_flg[5] = {NULL, OFF_CG, OFF_CG, OFF_CG, OFF_CG};

/* for shared memory */
#define SHMSZ	25		/* char 1byte * 18(22)sensor(degree) + light ... */
#define KEYCG	030700		/* 0x31c0 */
static key_t 	key[NUM_COM] = {KEYCG, KEYCG+1, KEYCG+2, KEYCG+3, KEYCG+4};
static int	shmid[NUM_COM];
static int	pid[NUM_COM];
static char	*shm[NUM_COM];

/* prototype of functions */
static int GetCgLoop(int cg_id);
static int Close_cgcg(int cg_id);
static int ReInit_cgcg(int cg_id);
static int ReStart_cgcg(int cg_id);
static int Request_cgcg(int cg_id);
static int Light_cgcg(int cg_id, int req);


int main(int argc, char *argv[])
{
	int i, cg_id, result = 0;

	if(argc != 4){
		fprintf(stderr, "usage: cgs: cgs cg-id, port-name baud-rate\n");
		exit(0);
	}

	cg_id = atoi(argv[1]);

	if(port_ref[cg_id] != -1){
		fprintf(stderr, "CyberGlove(%d) has been initialized\n", cg_id);
		return result;
	}

	result = OpenSP_com(cg_id, argv[2], atol(argv[3]));

	if(result != 1){
		exit(result);
	}

	ioctl(port_ref[cg_id], TCFLSH, 2); /* 2 means flush both buffers */
#if _DEBUG_
	fprintf(stderr, "open CG(%d) in childe process\n", cg_id);
#endif
	ReInit_cgcg(cg_id);
	ReStart_cgcg(cg_id);

	if((shmid[cg_id] = shmget(key[cg_id], SHMSZ, IPC_CREAT | 0666)) < 0){
		perror("shmget(cg-c)");
		exit(1);
	}

	if((shm[cg_id] = shmat(shmid[cg_id], NULL, 0)) == (char *) -1){
		perror("shmat(cg-c)");
		exit(1);
	}
#if _DEBUG_
	fprintf(stderr, "memmory sheared in childe process\n");
#endif
	GetCgLoop(cg_id);
	exit(0);
}


/*****     private functions...     *****/


static int GetCgLoop(int cg_id)
{
	char *shmp;
#if _DEBUG_
	unsigned long c;
#endif

	shmp = shm[cg_id];
	*shmp = 'd'; /* 'd' is data. */

	while(1){
#if _DEBUG_
	    for(c = 0; c < 9999; c++){
#endif
		if(*shmp == 'c'){ /* 'c' is command. */
			switch(*(shmp + 1)){
			    case 'c':
				Close_cgcg(cg_id);
#if _DEBUG_
				fprintf(stderr, "childe process is finished(cg%d)\n", cg_id);
#endif
				exit(0);
				break;
			    case 'i':
				ReInit_cgcg(cg_id);
				break;
			    case 's':
				ReStart_cgcg(cg_id);
				break;
			    case 'f':
				lit_flg[cg_id] = OFF_CG;
				Light_cgcg(cg_id, OFF_CG);
				break;
			    case 'n':
				lit_flg[cg_id] = OFF_CG;
				Light_cgcg(cg_id, ON_CG);
				break;
			    case 'l':
				lit_flg[cg_id] = ON_CG;
				Light_cgcg(cg_id, LIT_CG);
				break;
			    default:
				fprintf(stderr, "GetCgLoop: bad command %d\n", *(shmp +1));
				break;
			}
			*shmp = 'd';
		}
		Request_cgcg(cg_id);
	    }
#if _DEBUG_
	    fprintf(stderr, "GetCgLoop: childe process is running\n");
#endif
	}
	exit(1);
}


static int Close_cgcg(int cg_id)
{
	if(port_ref[cg_id] != -1){
		ioctl(port_ref[cg_id], TCSETA, &old_setup[cg_id]);
		close(port_ref[cg_id]);
		port_ref[cg_id] = -1;
		shmdt(shm[cg_id]);
		shmctl(shmid[cg_id], IPC_RMID);
		return 1;
	} else {
		fprintf(stderr, "cg(%d) has not been opened\n", cg_id);
		return 0;
	}
}


static int ReInit_cgcg(int cg_id)
{
	char c = CTL_COM('I');
	int result, i;

	result = (write(port_ref[cg_id], &c, 1) == 1);
	fprintf(stderr, "ReInit_cg: now re-initializing cg(%d), please wait 2 sec", cg_id);
	/* sleep(2); */
	for(i = 0; i < 2; i ++){
		sleep(1);
		fprintf(stderr, ".");
	}

	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != CTL_COM('I')){
		fprintf(stderr, "ReInit_cg(%d)-1: %d.\n", cg_id, c);
	}
#endif
	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != CR_COM){
		fprintf(stderr, "ReInit_cg(%d)-2: %d.\n", cg_id, c);
	}
#endif
	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != LF_COM){
		fprintf(stderr, "ReInit_cg(%d)-3: %d.\n", cg_id, c);
	}
#endif
	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != NUL_COM){
		fprintf(stderr, "ReInit_cg(%d)-4: %d.\n", cg_id, c);
	}
#endif

	fprintf(stderr, "done.\n");

	return result;
}


static int ReStart_cgcg(int cg_id)
{
	char c = CTL_COM('R');
	int result, i;

	result = (write(port_ref[cg_id], &c, 1) == 1);
	fprintf(stderr, "ReStart_cg: now re-starting cg(%d), please wait 3 sec", cg_id);
	/* sleep(3); */
	for(i = 0; i < 3; i ++){
		sleep(1);
		fprintf(stderr, ".");
	}

	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != CTL_COM('R')){
		fprintf(stderr, "ReInit_cg(%d)-1: %d.\n", cg_id, c);
	}
#endif
	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != CR_COM){
		fprintf(stderr, "ReInit_cg(%d)-2: %d.\n", cg_id, c);
	}
#endif
	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != LF_COM){
		fprintf(stderr, "ReInit_cg(%d)-2: %d.\n", cg_id, c);
	}
#endif
	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != NUL_COM){
		fprintf(stderr, "ReInit_cg(%d)-3: %d.\n", cg_id, c);
	}
#endif

	fprintf(stderr, "done.\n");

	return result;
}


static int Request_cgcg(int cg_id)
{
	int i, result;
	char c, *a;

	a = shm[cg_id] + 3;
	result = (write(port_ref[cg_id], "G", 1) == 1);

	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != 'G'){			/*   G */
		fprintf(stderr, "Request_cg(%d)-1: %d.\n", cg_id, c);
	}
#endif

	for(i = 0; i < 18; i++){
		read(port_ref[cg_id], a + i, 1);
	}
	/* read(port_cg[cg_id], a, 18);*/

	read(port_ref[cg_id], &c, 1);
#if _DEBUG_
	if(c != NUL_COM){
		fprintf(stderr, "Request_cg(%d)-2: %d.\n", cg_id, c);
	}
#endif

	/* get light */
	if(lit_flg[cg_id] == LIT_CG){
		result = (write(port_ref[cg_id], "?L", 2) == 2);
		read(port_ref[cg_id], &c, 1);
#if _DEBUG_
		if(c != '?'){
			fprintf(stderr, "Light_cg(%d)-r1: %d.\n", cg_id, c);
		}
#endif
		read(port_ref[cg_id], &c, 1);
#if _DEBUG_
		if(c != 'L'){
			fprintf(stderr, "Light_cg(%d)-r2: %d.\n", cg_id, c);
		}
#endif
		read(port_ref[cg_id], &c, 1);
		result = (int) c;
		*(a - 1) = c;			/* **L18... */
		read(port_ref[cg_id], &c, 1);
#if _DEBUG_
		if(c != NUL_COM){
			fprintf(stderr, "Light_cg(%d)-r3: %d.\n", cg_id, c);
		}
#endif
	}

	ioctl(port_ref[cg_id], TCFLSH, 2); /* 2 means flush both buffers */

	return result;
}


static int Light_cgcg(int cg_id, int req)
{
	int result;
	char c, s[2] = "L0";

	switch(req){
	    case OFF_CG:
	    case ON_CG:
		*(shm[cg_id] + 2) = (char) req;		/* **L18... */
		s[1] = (char) req;
		result = (write(port_ref[cg_id], s, 2) == 2);
		read(port_ref[cg_id], &c, 1);
#if _DEBUG_
		if(c != 'L'){
			fprintf(stderr, "Light_cg(%d)-1: %d(%d).\n", cg_id, c);
		}
#endif
		read(port_ref[cg_id], &c, 1);
#if _DEBUG_
		if(c != NUL_COM){
			fprintf(stderr, "Light_cg(%d)-2: %d.\n", cg_id, c);
		}
#endif
		break;
	    case LIT_CG:
		s[0] = 'J';
		s[1] = 1;
		result = (write(port_ref[cg_id], s, 2) == 2);
		read(port_ref[cg_id], &c, 1);
#if _DEBUG_
		if(c != 'J'){
			fprintf(stderr, "Light_cg(%d)-3: %d.\n", cg_id, c);
		}
#endif
		read(port_ref[cg_id], &c, 1);
#if _DEBUG_
		if(c != NUL_COM){
			fprintf(stderr, "Light_cg(%d)-4: %d.\n", cg_id, c);
		}
#endif
		break;
	}

	return result;
}
